@@ -85,33 +85,83 @@ ByteCount get_byte_to_column(const Buffer& buffer, ColumnCount tabstop, DisplayC
     return (int)(it - line.begin());
 }
 
+static BufferLines parse_lines(const char* pos, const char* end, EolFormat eolformat)
+{
+    BufferLines lines;
+    while (pos < end)
+    {
+        if (lines.size() >= std::numeric_limits<int>::max())
+            throw runtime_error("too many lines");
+
+        const char* eol = std::find(pos, end, '\n');
+        if ((eol - pos) >= std::numeric_limits<int>::max())
+            throw runtime_error("line is too long");
+
+        lines.emplace_back(StringData::create(StringView{pos, eol - (eolformat == EolFormat::Crlf and eol != end ? 1 : 0)}, "\n"));
+        pos = eol + 1;
+    }
+
+    if (lines.empty())
+        lines.emplace_back(StringData::create("\n"));
+
+    return lines;
+}
+
+Buffer* create_buffer_from_string(String name, Buffer::Flags flags, StringView data)
+{
+    return BufferManager::instance().create_buffer(
+        std::move(name), flags,
+        parse_lines(data.begin(), data.end(), EolFormat::Lf),
+        ByteOrderMark::None, EolFormat::Lf,
+        FsStatus{InvalidTime, {}, {}});
+}
+
+template<typename Func>
+decltype(auto) parse_file(StringView filename, Func&& func)
+{
+    MappedFile file{parse_filename(filename)};
+
+    const char* pos = file.data;
+    const char* end = pos + file.st.st_size;
+
+    auto bom = ByteOrderMark::None;
+    if (file.st.st_size >= 3 && StringView{pos, 3_byte} == "\xEF\xBB\xBF")
+    {
+        bom = ByteOrderMark::Utf8;
+        pos += 3;
+    }
+
+    bool has_crlf = false, has_lf = false;
+    for (auto it = std::find(pos, end, '\n'); it != end; it = std::find(it+1, end, '\n'))
+        ((it != pos and *(it-1) == '\r') ? has_crlf : has_lf) = true;
+    auto eolformat = (has_crlf and not has_lf) ? EolFormat::Crlf : EolFormat::Lf;
+
+    FsStatus fs_status{file.st.st_mtim, file.st.st_size, murmur3(file.data, file.st.st_size)};
+    return func(parse_lines(pos, end, eolformat), bom, eolformat, fs_status);
+}
+
 Buffer* open_file_buffer(StringView filename, Buffer::Flags flags)
 {
-    MappedFile file_data{parse_filename(filename)};
-    return BufferManager::instance().create_buffer(
-        filename.str(), Buffer::Flags::File | flags, file_data, file_data.st.st_mtim);
+    return parse_file(filename, [&](BufferLines&& lines, ByteOrderMark bom, EolFormat eolformat, FsStatus fs_status)  {
+        return BufferManager::instance().create_buffer(filename.str(), Buffer::Flags::File | flags,
+                                                       std::move(lines), bom, eolformat, fs_status);
+    });
 }
 
 Buffer* open_or_create_file_buffer(StringView filename, Buffer::Flags flags)
 {
-    auto& buffer_manager = BufferManager::instance();
     auto path = parse_filename(filename);
     if (file_exists(path))
-    {
-        MappedFile file_data{path};
-        return buffer_manager.create_buffer(filename.str(), Buffer::Flags::File | flags,
-                                            file_data, file_data.st.st_mtim);
-    }
-    return buffer_manager.create_buffer(
-        filename.str(), Buffer::Flags::File | Buffer::Flags::New,
-        {}, InvalidTime);
+        return open_file_buffer(filename.str(), Buffer::Flags::File | flags);
+    return create_buffer_from_string(filename.str(), Buffer::Flags::File | Buffer::Flags::New, StringView{});
 }
 
 void reload_file_buffer(Buffer& buffer)
 {
     kak_assert(buffer.flags() & Buffer::Flags::File);
-    MappedFile file_data{buffer.name()};
-    buffer.reload(file_data, file_data.st.st_mtim);
+    parse_file(buffer.name(), [&](auto&&... params) {
+        buffer.reload(std::forward<decltype(params)>(params)...);
+    });
     buffer.flags() &= ~Buffer::Flags::New;
 }
 
@@ -124,16 +174,19 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
     if (buffer)
     {
         buffer->flags() |= Buffer::Flags::NoUndo | flags;
-        buffer->reload({}, InvalidTime);
+        buffer->values().clear();
+        buffer->reload({StringData::create("\n")}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
+        buffer_manager.make_latest(*buffer);
     }
     else
         buffer = buffer_manager.create_buffer(
-            std::move(name), flags | Buffer::Flags::Fifo | Buffer::Flags::NoUndo);
+            std::move(name), flags | Buffer::Flags::Fifo | Buffer::Flags::NoUndo,
+            {StringData::create("\n")}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
 
     struct FifoWatcher : FDWatcher
     {
         FifoWatcher(int fd, Buffer& buffer, bool scroll)
-            : FDWatcher(fd, FdEvents::Read,
+            : FDWatcher(fd, FdEvents::Read, EventMode::Normal,
                         [](FDWatcher& watcher, FdEvents, EventMode mode) {
                             if (mode == EventMode::Normal)
                                 static_cast<FifoWatcher&>(watcher).read_fifo();
@@ -149,7 +202,7 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
             m_buffer.flags() &= ~(Buffer::Flags::Fifo | Buffer::Flags::NoUndo);
         }
 
-        void read_fifo() const
+        void read_fifo()
         {
             kak_assert(m_buffer.flags() & Buffer::Flags::Fifo);
 
@@ -157,7 +210,7 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
             // if we read data slower than it arrives in the fifo, limiting the
             // iteration number allows us to go back go back to the event loop and
             // handle other events sources (such as input)
-            constexpr size_t max_loop = 16;
+            constexpr size_t max_loop = 1024;
             bool closed = false;
             size_t loop = 0;
             char data[buffer_size];
@@ -178,20 +231,21 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
                     }
 
                     auto pos = m_buffer.back_coord();
-                    const bool prevent_scrolling = pos == BufferCoord{0,0} and not m_scroll;
-                    if (prevent_scrolling)
+                    const bool is_first = pos == BufferCoord{0,0};
+                    if (not m_scroll and (is_first or m_had_trailing_newline))
                         pos = m_buffer.next(pos);
 
-                    m_buffer.insert(pos, StringView(data, data+count));
+                    pos = m_buffer.insert(pos, StringView(data, data+count)).end;
 
-                    if (prevent_scrolling)
+                    bool have_trailing_newline = (data[count-1] == '\n');
+                    if (not m_scroll)
                     {
-                        m_buffer.erase({0,0}, m_buffer.next({0,0}));
-                        // in the other case, the buffer will have automatically
-                        // inserted a \n to guarantee its invariant.
-                        if (data[count-1] == '\n')
-                            m_buffer.insert(m_buffer.end_coord(), "\n");
+                        if (is_first)
+                            m_buffer.erase({0,0}, m_buffer.next({0,0}));
+                        else if (not m_had_trailing_newline and have_trailing_newline)
+                            m_buffer.erase(m_buffer.prev(pos), pos);
                     }
+                    m_had_trailing_newline = have_trailing_newline;
                 }
                 while (++loop < max_loop  and fd_readable(fifo));
             }
@@ -207,9 +261,10 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
 
         Buffer& m_buffer;
         bool m_scroll;
+        bool m_had_trailing_newline = false;
     };
 
-    buffer->values()[fifo_watcher_id] = Value(std::make_unique<FifoWatcher>(fd, *buffer, scroll));
+    buffer->values()[fifo_watcher_id] = Value(Meta::Type<FifoWatcher>{}, fd, *buffer, scroll);
     buffer->flags() = flags | Buffer::Flags::Fifo | Buffer::Flags::NoUndo;
     buffer->run_hook_in_own_context(Hook::BufOpenFifo, buffer->name());
 
@@ -239,9 +294,9 @@ void write_to_debug_buffer(StringView str)
     else
     {
         String line = str + (eol_back ? "\n" : "\n\n");
-        BufferManager::instance().create_buffer(
+        create_buffer_from_string(
             debug_buffer_name.str(), Buffer::Flags::NoUndo | Buffer::Flags::Debug | Buffer::Flags::ReadOnly,
-            line, InvalidTime);
+            line);
     }
 }
 
