@@ -31,15 +31,18 @@ public:
     InputMode(const InputMode&) = delete;
     InputMode& operator=(const InputMode&) = delete;
 
-    void handle_key(Key key) { RefPtr<InputMode> keep_alive{this}; on_key(key); }
+    void handle_key(Key key, bool synthesized) { RefPtr<InputMode> keep_alive{this}; on_key(key, synthesized); }
+    virtual void paste(StringView content);
 
-    virtual void on_enabled() {}
-    virtual void on_disabled(bool temporary) {}
+    virtual void on_enabled(bool from_pop) {}
+    virtual void on_disabled(bool from_push) {}
+
+    virtual void refresh_ifn() {}
 
     bool enabled() const { return &m_input_handler.current_mode() == this; }
     Context& context() const { return m_input_handler.context(); }
 
-    virtual DisplayLine mode_line() const = 0;
+    virtual ModeInfo mode_info() const = 0;
 
     virtual KeymapMode keymap_mode() const = 0;
 
@@ -53,10 +56,9 @@ public:
     }
 
     using Insertion = InputHandler::Insertion;
-    Insertion& last_insert() { return m_input_handler.m_last_insert; }
 
 protected:
-    virtual void on_key(Key key) = 0;
+    virtual void on_key(Key key, bool synthesized) = 0;
 
     void push_mode(InputMode* new_mode)
     {
@@ -71,6 +73,32 @@ private:
     InputHandler& m_input_handler;
 };
 
+void InputMode::paste(StringView content)
+{
+    try
+    {
+        Buffer& buffer = context().buffer();
+        const bool linewise = not content.empty() and content.back() == '\n';
+        ScopedEdition edition{context()};
+        ScopedSelectionEdition selection_edition{context()};
+        context().selections().for_each([&buffer, content=std::move(content), linewise]
+                                        (size_t index, Selection& sel) {
+            auto& min = sel.min();
+            auto& max = sel.max();
+            BufferRange range =
+                buffer.insert(paste_pos(buffer, min, max, PasteMode::Insert, linewise), content);
+            min = range.begin;
+            max = range.end > range.begin ? buffer.char_prev(range.end) : range.begin;
+        }, false);
+    }
+    catch (Kakoune::runtime_error& error)
+    {
+        write_to_debug_buffer(format("Error: {}", error.what()));
+        context().print_status({error.what().str(), context().faces()["Error"] });
+        context().hooks().run_hook(Hook::RuntimeError, error.what(), context());
+    }
+}
+
 namespace InputModes
 {
 
@@ -93,59 +121,73 @@ struct MouseHandler
 
         Buffer& buffer = context.buffer();
         BufferCoord cursor;
-        auto& selections = context.selections();
         constexpr auto modifiers = Key::Modifiers::Control | Key::Modifiers::Alt | Key::Modifiers::Shift | Key::Modifiers::MouseButtonMask;
         switch ((key.modifiers & ~modifiers).value)
         {
         case Key::Modifiers::MousePress:
             switch (key.mouse_button())
             {
-            case Key::MouseButton::Right:
-                m_dragging = false;
+            case Key::MouseButton::Right: {
+                m_dragging.reset();
                 cursor = context.window().buffer_coord(key.coord());
+                ScopedSelectionEdition selection_edition{context};
+                auto& selections = context.selections();
                 if (key.modifiers & Key::Modifiers::Control)
                     selections = {{selections.begin()->anchor(), cursor}};
                 else
                     selections.main() = {selections.main().anchor(), cursor};
                 selections.sort_and_merge_overlapping();
                 return true;
+            }
 
-            case Key::MouseButton::Left:
-                m_dragging = true;
+            case Key::MouseButton::Left: {
+                m_dragging.reset(new ScopedSelectionEdition{context});
                 m_anchor = context.window().buffer_coord(key.coord());
                 if (not (key.modifiers & Key::Modifiers::Control))
                     context.selections_write_only() = { buffer, m_anchor};
                 else
                 {
+                    auto& selections = context.selections();
                     size_t main = selections.size();
                     selections.push_back({m_anchor});
                     selections.set_main_index(main);
                     selections.sort_and_merge_overlapping();
                 }
                 return true;
+            }
 
             default: return true;
             }
 
-        case Key::Modifiers::MouseRelease:
+        case Key::Modifiers::MouseRelease: {
             if (not m_dragging)
+            {
+                context.ensure_cursor_visible = false;
                 return true;
-            m_dragging = false;
+            }
+            auto& selections = context.selections();
             cursor = context.window().buffer_coord(key.coord());
             selections.main() = {buffer.clamp(m_anchor), cursor};
             selections.sort_and_merge_overlapping();
+            m_dragging.reset();
             return true;
+        }
 
-        case Key::Modifiers::MousePos:
+        case Key::Modifiers::MousePos: {
             if (not m_dragging)
+            {
+                context.ensure_cursor_visible = false;
                 return true;
+            }
             cursor = context.window().buffer_coord(key.coord());
+            auto& selections = context.selections();
             selections.main() = {buffer.clamp(m_anchor), cursor};
             selections.sort_and_merge_overlapping();
             return true;
+        }
 
         case Key::Modifiers::Scroll:
-            scroll_window(context, static_cast<int32_t>(key.key), m_dragging);
+            scroll_window(context, static_cast<int32_t>(key.key), m_dragging ? OnHiddenCursor::MoveCursor : OnHiddenCursor::PreserveSelections);
             return true;
 
         default: return false;
@@ -153,7 +195,7 @@ struct MouseHandler
     }
 
 private:
-    bool m_dragging = false;
+    std::unique_ptr<ScopedSelectionEdition> m_dragging;
     BufferCoord m_anchor;
 };
 
@@ -192,7 +234,7 @@ public:
           m_state(single_command ? State::SingleCommand : State::Normal)
     {}
 
-    void on_enabled() override
+    void on_enabled(bool from_pop) override
     {
         if (m_state == State::PopOnEnabled)
             return pop_mode();
@@ -213,19 +255,19 @@ public:
         }
     }
 
-    void on_disabled(bool temporary) override
+    void on_disabled(bool from_push) override
     {
         m_idle_timer.disable();
         m_fs_check_timer.disable();
 
-        if (not temporary and m_hooks_disabled)
+        if (not from_push and m_hooks_disabled)
         {
             context().hooks_disabled().unset();
             m_hooks_disabled = false;
         }
     }
 
-    void on_key(Key key) override
+    void on_key(Key key, bool) override
     {
         kak_assert(m_state != State::PopOnEnabled);
         ScopedSetBool set_in_on_key{m_in_on_key};
@@ -328,7 +370,7 @@ public:
             m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
     }
 
-    DisplayLine mode_line() const override
+    ModeInfo mode_info() const override
     {
         AtomList atoms;
         auto num_sel = context().selections().size();
@@ -348,7 +390,7 @@ public:
             atoms.emplace_back(" reg=", context().faces()["StatusLineInfo"]);
             atoms.emplace_back(StringView(m_params.reg).str(), context().faces()["StatusLineValue"]);
         }
-        return atoms;
+        return {atoms, m_params};
     }
 
     KeymapMode keymap_mode() const override { return KeymapMode::Normal; }
@@ -595,143 +637,6 @@ private:
     const FaceRegistry& m_faces;
 };
 
-class Menu : public InputMode
-{
-public:
-    Menu(InputHandler& input_handler, Vector<DisplayLine> choices,
-         MenuCallback callback)
-        : InputMode(input_handler),
-          m_callback(std::move(callback)), m_choices(choices.begin(), choices.end()),
-          m_selected(m_choices.begin()),
-          m_filter_editor{context().faces()}
-    {
-        if (not context().has_client())
-            return;
-        context().client().menu_show(std::move(choices), {}, MenuStyle::Prompt);
-        context().client().menu_select(0);
-    }
-
-    void on_key(Key key) override
-    {
-        auto match_filter = [this](const DisplayLine& choice) {
-            for (auto& atom : choice)
-            {
-                const auto& contents = atom.content();
-                if (regex_match(contents.begin(), contents.end(), m_filter))
-                    return true;
-            }
-            return false;
-        };
-
-        if (key == Key::Return)
-        {
-            if (context().has_client())
-                context().client().menu_hide();
-            context().print_status(DisplayLine{});
-
-            // Maintain hooks disabled in callback if they were before pop_mode
-            ScopedSetBool disable_hooks(context().hooks_disabled(),
-                                        context().hooks_disabled());
-            pop_mode();
-            int selected = m_selected - m_choices.begin();
-            m_callback(selected, MenuEvent::Validate, context());
-            return;
-        }
-        else if (key == Key::Escape or key == ctrl('c'))
-        {
-            if (m_edit_filter)
-            {
-                m_edit_filter = false;
-                m_filter = Regex{".*"};
-                m_filter_editor.reset("", "");
-                context().print_status(DisplayLine{});
-            }
-            else
-            {
-                if (context().has_client())
-                    context().client().menu_hide();
-
-                // Maintain hooks disabled in callback if they were before pop_mode
-                ScopedSetBool disable_hooks(context().hooks_disabled(),
-                                            context().hooks_disabled());
-                pop_mode();
-                int selected = m_selected - m_choices.begin();
-                m_callback(selected, MenuEvent::Abort, context());
-            }
-        }
-        else if (key == Key::Down or key == Key::Tab or
-                 key == ctrl('n') or (not m_edit_filter and key == 'j'))
-        {
-            auto it = std::find_if(m_selected+1, m_choices.end(), match_filter);
-            if (it == m_choices.end())
-                it = std::find_if(m_choices.begin(), m_selected, match_filter);
-            select(it);
-        }
-        else if (key == Key::Up or key == shift(Key::Tab) or
-                 key == ctrl('p') or (not m_edit_filter and key == 'k'))
-        {
-            ChoiceList::const_reverse_iterator selected(m_selected+1);
-            auto it = std::find_if(selected+1, m_choices.rend(), match_filter);
-            if (it == m_choices.rend())
-                it = std::find_if(m_choices.rbegin(), selected, match_filter);
-            select(it.base()-1);
-        }
-        else if (key == '/' and not m_edit_filter)
-        {
-            m_edit_filter = true;
-        }
-        else if (m_edit_filter)
-        {
-            m_filter_editor.handle_key(key);
-
-            auto search = ".*" + m_filter_editor.line() + ".*";
-            m_filter = Regex{search};
-            auto it = std::find_if(m_selected, m_choices.end(), match_filter);
-            if (it == m_choices.end())
-                it = std::find_if(m_choices.begin(), m_selected, match_filter);
-            select(it);
-        }
-
-        if (m_edit_filter and context().has_client())
-        {
-            auto prompt = "filter:"_str;
-            auto width = context().client().dimensions().column - prompt.column_length();
-            auto display_line = m_filter_editor.build_display_line(width);
-            display_line.insert(display_line.begin(), { prompt, context().faces()["Prompt"] });
-            context().print_status(display_line);
-        }
-    }
-
-    DisplayLine mode_line() const override
-    {
-        return { "menu", context().faces()["StatusLineMode"] };
-    }
-
-    KeymapMode keymap_mode() const override { return KeymapMode::Menu; }
-
-    StringView name() const override { return "menu"; }
-
-private:
-    MenuCallback m_callback;
-
-    using ChoiceList = Vector<DisplayLine>;
-    const ChoiceList m_choices;
-    ChoiceList::const_iterator m_selected;
-
-    void select(ChoiceList::const_iterator it)
-    {
-        m_selected = it;
-        int selected = m_selected - m_choices.begin();
-        if (context().has_client())
-            context().client().menu_select(selected);
-        m_callback(selected, MenuEvent::Select, context());
-    }
-
-    Regex      m_filter = Regex{".*"};
-    bool       m_edit_filter = false;
-    LineEditor m_filter_editor;
-};
-
 static Optional<Codepoint> get_raw_codepoint(Key key)
 {
     if (auto cp = key.codepoint())
@@ -752,7 +657,9 @@ public:
         : InputMode(input_handler), m_callback(std::move(callback)), m_completer(std::move(completer)),
           m_prompt(prompt.str()), m_prompt_face(face),
           m_empty_text{std::move(emptystr)},
-          m_line_editor{context().faces()}, m_flags(flags),
+          // This prompt may outlive local scopes so ignore local faces.
+          m_line_editor{context().faces(false)}, m_flags(flags),
+          m_was_interactive{not context().noninteractive()},
           m_history{RegisterManager::instance()[history_register]},
           m_current_history{-1},
           m_auto_complete{context().options()["autocomplete"].get<AutoComplete>() & AutoComplete::Prompt},
@@ -772,17 +679,20 @@ public:
         m_line_editor.reset(std::move(initstr), m_empty_text);
     }
 
-    void on_key(Key key) override
+    void on_key(Key key, bool synthesized) override
     {
         const String& line = m_line_editor.line();
+        if (not synthesized)
+            m_was_interactive = true;
 
         auto can_auto_insert_completion = [&] {
             const bool has_completions = not m_completions.candidates.empty();
             const bool completion_selected = m_current_completion != -1;
             const bool text_entered = m_completions.start != line.byte_count_to(m_line_editor.cursor_pos());
+            const bool at_end = line.byte_count_to(m_line_editor.cursor_pos()) == line.length();
             return (m_completions.flags & Completions::Flags::Menu) and
                 has_completions and
-                not completion_selected and
+                not completion_selected and at_end and
                 (not (m_completions.flags & Completions::Flags::NoEmpty) or text_entered);
         };
 
@@ -892,8 +802,10 @@ public:
         else if (key == Key::Tab or key == shift(Key::Tab) or key.modifiers == Key::Modifiers::MenuSelect) // completion
         {
             CandidateList& candidates = m_completions.candidates;
-            // first try, we need to ask our completer for completions
-            if (candidates.empty())
+
+            if (m_auto_complete and m_refresh_completion_pending)
+                refresh_completions(CompletionFlags::Fast);
+            if (candidates.empty()) // manual completion, we need to ask our completer for completions
             {
                 refresh_completions(CompletionFlags::None);
                 if ((not m_prefix_in_completions and candidates.size() > 1) or
@@ -933,7 +845,7 @@ public:
         {
             on_next_key_with_autoinfo(context(), "explicit-completion", KeymapMode::None,
                 [this](Key key, Context&) {
-                    m_explicit_completer = PromptCompleter{};
+                    m_explicit_completer = ArgCompleter{};
 
                     if (key.key == 'f')
                         use_explicit_completer([](const Context& context, StringView token) {
@@ -960,7 +872,7 @@ public:
         }
         else if (key == ctrl('o'))
         {
-            m_explicit_completer = PromptCompleter{};
+            m_explicit_completer = ArgCompleter{};
             m_auto_complete = not m_auto_complete;
 
             if (m_auto_complete)
@@ -1007,6 +919,30 @@ public:
             m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
     }
 
+    void refresh_ifn() override
+    {
+        bool explicit_completion_selected = m_current_completion != -1 and
+            (not m_prefix_in_completions or m_current_completion != m_completions.candidates.size() - 1);
+        if (not enabled() or (context().flags() & Context::Flags::Draft) or explicit_completion_selected)
+            return;
+
+        if (auto next_date = Clock::now() + get_idle_timeout(context());
+            next_date < m_idle_timer.next_date())
+            m_idle_timer.set_next_date(next_date);
+        m_refresh_completion_pending = true;
+    }
+
+    void paste(StringView content) override
+    {
+        m_line_editor.insert(content);
+        clear_completions();
+        m_refresh_completion_pending = true;
+        display();
+        m_line_changed = true;
+        if (not (context().flags() & Context::Flags::Draft))
+            m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
+    }
+
     void set_prompt_face(Face face)
     {
         if (face != m_prompt_face)
@@ -1016,9 +952,9 @@ public:
         }
     }
 
-    DisplayLine mode_line() const override
+    ModeInfo mode_info() const override
     {
-        return { "prompt", context().faces()["StatusLineMode"] };
+        return {{ "prompt", context().faces()["StatusLineMode"] }, {}};
     }
 
     KeymapMode keymap_mode() const override { return KeymapMode::Prompt; }
@@ -1063,35 +999,35 @@ private:
             const String& line = m_line_editor.line();
             m_completions = completer(context(), flags, line,
                                       line.byte_count_to(m_line_editor.cursor_pos()));
+            if (not context().has_client())
+                return;
+            if (m_completions.candidates.empty())
+                return context().client().menu_hide();
+
+            show_completions();
             const bool menu = (bool)(m_completions.flags & Completions::Flags::Menu);
-            if (context().has_client())
+            if (menu)
+                context().client().menu_select(0);
+            auto prefix = line.substr(m_completions.start, m_completions.end - m_completions.start);
+            m_prefix_in_completions = not menu and not contains(m_completions.candidates, prefix);
+            if (m_prefix_in_completions)
             {
-                if (m_completions.candidates.empty())
-                    return context().client().menu_hide();
-
-                Vector<DisplayLine> items;
-                for (auto& candidate : m_completions.candidates)
-                    items.push_back({ candidate, {} });
-
-                const auto menu_style = (m_flags & PromptFlags::Search) ? MenuStyle::Search : MenuStyle::Prompt;
-                context().client().menu_show(items, {}, menu_style);
-
-                if (menu)
-                    context().client().menu_select(0);
-
-                auto prefix = line.substr(m_completions.start, m_completions.end - m_completions.start);
-                if (not menu and not contains(m_completions.candidates, prefix))
-                {
-                    m_current_completion = m_completions.candidates.size();
-                    m_completions.candidates.push_back(prefix.str());
-                    m_prefix_in_completions = true;
-                }
-                else
-                    m_prefix_in_completions = false;
+                m_current_completion = m_completions.candidates.size();
+                m_completions.candidates.push_back(prefix.str());
             }
         } catch (runtime_error&) {}
     }
 
+    void show_completions()
+    {
+        Vector<DisplayLine> items;
+        for (auto& candidate : m_completions.candidates)
+            items.push_back({ candidate, {} });
+
+        const auto menu_style = (m_flags & PromptFlags::Search) ? MenuStyle::Search : MenuStyle::Prompt;
+        context().client().menu_show(std::move(items), {}, menu_style);
+    }
+
     void clear_completions()
     {
         m_current_completion = -1;
@@ -1111,18 +1047,31 @@ private:
         context().print_status(display_line);
     }
 
-    void on_enabled() override
+    void on_enabled(bool from_pop) override
     {
         display();
-        m_line_changed = true;
+        if (from_pop)
+        {
+            if (context().has_client() and not m_completions.candidates.empty())
+            {
+                show_completions();
+                const bool menu = (bool)(m_completions.flags & Completions::Flags::Menu);
+                if (m_current_completion != -1)
+                    context().client().menu_select(m_current_completion);
+                else if (menu)
+                    context().client().menu_select(0);
+            }
+        }
+        else
+            m_line_changed = true;
 
         if (not (context().flags() & Context::Flags::Draft))
             m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
     }
 
-    void on_disabled(bool temporary) override
+    void on_disabled(bool from_push) override
     {
-        if (not temporary)
+        if (not from_push)
             context().print_status({});
 
         m_idle_timer.disable();
@@ -1143,6 +1092,7 @@ private:
     LineEditor     m_line_editor;
     bool           m_line_changed = false;
     PromptFlags    m_flags;
+    bool           m_was_interactive;
     Register&      m_history;
     int            m_current_history;
     bool           m_auto_complete;
@@ -1151,7 +1101,7 @@ private:
 
     void history_push(StringView entry)
     {
-        if (entry.empty() or context().history_disabled() or
+        if (entry.empty() or not m_was_interactive or
             (m_flags & PromptFlags::DropHistoryEntriesWithBlankPrefix and
              is_horizontal_blank(entry[0_byte])))
             return;
@@ -1168,7 +1118,7 @@ public:
         : InputMode(input_handler), m_name{std::move(name)}, m_callback(std::move(callback)), m_keymap_mode(keymap_mode),
           m_idle_timer{Clock::now() + get_idle_timeout(context()), std::move(idle_callback)} {}
 
-    void on_key(Key key) override
+    void on_key(Key key, bool) override
     {
         // maintain hooks disabled in the callback if they were before pop_mode
         ScopedSetBool disable_hooks(context().hooks_disabled(),
@@ -1177,9 +1127,9 @@ public:
         m_callback(key, context());
     }
 
-    DisplayLine mode_line() const override
+    ModeInfo mode_info() const override
     {
-        return { "enter key", context().faces()["StatusLineMode"] };
+        return {{ "enter key", context().faces()["StatusLineMode"] }, {}};
     }
 
     KeymapMode keymap_mode() const override { return m_keymap_mode; }
@@ -1196,10 +1146,12 @@ private:
 class Insert : public InputMode
 {
 public:
-    Insert(InputHandler& input_handler, InsertMode mode, int count)
+    Insert(InputHandler& input_handler, InsertMode mode, int count, Insertion* last_insert)
         : InputMode(input_handler),
           m_edition(context()),
+          m_selection_edition(context()),
           m_completer(context()),
+          m_last_insert(last_insert),
           m_restore_cursor(mode == InsertMode::Append),
           m_auto_complete{context().options()["autocomplete"].get<AutoComplete>() & AutoComplete::Insert},
           m_idle_timer{TimePoint::max(), context().flags() & Context::Flags::Draft ?
@@ -1212,27 +1164,31 @@ public:
     {
         context().buffer().throw_if_read_only();
 
-        last_insert().recording.set();
-        last_insert().mode = mode;
-        last_insert().keys.clear();
-        last_insert().disable_hooks = context().hooks_disabled();
-        last_insert().count = count;
+        if (m_last_insert)
+        {
+            m_last_insert->recording.set();
+            m_last_insert->mode = mode;
+            m_last_insert->keys.clear();
+            m_last_insert->disable_hooks = context().hooks_disabled();
+            m_last_insert->count = count;
+        }
         prepare(mode, count);
     }
 
-    void on_enabled() override
+    void on_enabled(bool from_pop) override
     {
         if (not (context().flags() & Context::Flags::Draft))
             m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
     }
 
-    void on_disabled(bool temporary) override
+    void on_disabled(bool from_push) override
     {
         m_idle_timer.disable();
 
-        if (not temporary)
+        if (not from_push)
         {
-            last_insert().recording.unset();
+            if (m_last_insert)
+                m_last_insert->recording.unset();
 
             auto& selections = context().selections();
             if (m_restore_cursor)
@@ -1246,7 +1202,7 @@ public:
         }
     }
 
-    void on_key(Key key) override
+    void on_key(Key key, bool synthesized) override
     {
         auto& buffer = context().buffer();
 
@@ -1344,22 +1300,13 @@ public:
                 }, "enter register name", register_doc.str());
             update_completions = false;
         }
-        else if (key == ctrl('n'))
+        else if (key == ctrl('n') or key == ctrl('p') or key.modifiers == Key::Modifiers::MenuSelect)
         {
-            last_insert().keys.pop_back();
-            m_completer.select(1, true, last_insert().keys);
-            update_completions = false;
-        }
-        else if (key == ctrl('p'))
-        {
-            last_insert().keys.pop_back();
-            m_completer.select(-1, true, last_insert().keys);
-            update_completions = false;
-        }
-        else if (key.modifiers == Key::Modifiers::MenuSelect)
-        {
-            last_insert().keys.pop_back();
-            m_completer.select(key.key, false, last_insert().keys);
+            if (m_last_insert and not synthesized)
+                m_last_insert->keys.pop_back();
+            bool relative = key.modifiers != Key::Modifiers::MenuSelect;
+            int index = relative ? (key == ctrl('n') ? 1 : -1) : key.key;
+            m_completer.select(index, relative, m_last_insert ? &m_last_insert->keys : nullptr);
             update_completions = false;
         }
         else if (key == ctrl('x'))
@@ -1424,15 +1371,22 @@ public:
             m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
     }
 
-    DisplayLine mode_line() const override
+    void paste(StringView content) override
+    {
+        insert(ConstArrayView<StringView>{content});
+        m_idle_timer.set_next_date(Clock::now() + get_idle_timeout(context()));
+    }
+
+    ModeInfo mode_info() const override
     {
         auto num_sel = context().selections().size();
         auto main_index = context().selections().main_index();
-        return {AtomList{ { "insert", context().faces()["StatusLineMode"] },
-                          { " ", context().faces()["StatusLine"] },
-                          { num_sel == 1 ? format("{} sel", num_sel)
-                              : format("{} sels ({})", num_sel, main_index + 1),
-                            context().faces()["StatusLineInfo"] } }};
+        return {{AtomList{ { "insert", context().faces()["StatusLineMode"] },
+                           { " ", context().faces()["StatusLine"] },
+                           { num_sel == 1 ? format("{} sel", num_sel)
+                               : format("{} sels ({})", num_sel, main_index + 1),
+                             context().faces()["StatusLineInfo"] } }},
+                {}};
     }
 
     KeymapMode keymap_mode() const override { return KeymapMode::Insert; }
@@ -1453,19 +1407,20 @@ private:
         selections.sort_and_merge_overlapping();
     }
 
-    void insert(ConstArrayView<String> strings)
+    template<typename S>
+    void insert(ConstArrayView<S> strings)
     {
         m_completer.try_accept();
         context().selections().for_each([strings, &buffer=context().buffer()]
                                         (size_t index, Selection& sel) {
             Kakoune::insert(buffer, sel, sel.cursor(), strings[std::min(strings.size()-1, index)]);
-        });
+        }, false);
     }
 
     void insert(Codepoint key)
     {
         String str{key};
-        insert(str);
+        insert(ConstArrayView<String>{str});
         context().hooks().run_hook(Hook::InsertChar, str, context());
     }
 
@@ -1549,7 +1504,9 @@ private:
     }
 
     ScopedEdition   m_edition;
+    ScopedSelectionEdition m_selection_edition;
     InsertCompleter m_completer;
+    Insertion*      m_last_insert;
     const bool      m_restore_cursor;
     bool            m_auto_complete;
     Timer           m_idle_timer;
@@ -1563,7 +1520,7 @@ InputHandler::InputHandler(SelectionList selections, Context::Flags flags, Strin
     : m_context(*this, std::move(selections), flags, std::move(name))
 {
     m_mode_stack.emplace_back(new InputModes::Normal(*this));
-    current_mode().on_enabled();
+    current_mode().on_enabled(false);
 }
 
 InputHandler::~InputHandler() = default;
@@ -1574,7 +1531,7 @@ void InputHandler::push_mode(InputMode* new_mode)
 
     current_mode().on_disabled(true);
     m_mode_stack.emplace_back(new_mode);
-    new_mode->on_enabled();
+    new_mode->on_enabled(false);
 
     context().hooks().run_hook(Hook::ModeChange, format("push:{}:{}", prev_name, new_mode->name()), context());
 }
@@ -1589,7 +1546,7 @@ void InputHandler::pop_mode(InputMode* mode)
 
     current_mode().on_disabled(false);
     m_mode_stack.pop_back();
-    current_mode().on_enabled();
+    current_mode().on_enabled(true);
 
     context().hooks().run_hook(Hook::ModeChange, format("pop:{}:{}", prev_name, current_mode().name()), context());
 }
@@ -1606,7 +1563,7 @@ void InputHandler::reset_normal_mode()
 
 void InputHandler::insert(InsertMode mode, int count)
 {
-    push_mode(new InputModes::Insert(*this, mode, count));
+    push_mode(new InputModes::Insert(*this, mode, count, m_handle_key_level <= 1 ? &m_last_insert : nullptr));
 }
 
 void InputHandler::repeat_last_insert()
@@ -1618,22 +1575,20 @@ void InputHandler::repeat_last_insert()
         m_last_insert.recording)
         throw runtime_error{"repeating last insert not available in this context"};
 
-    Vector<Key> keys;
-    swap(keys, m_last_insert.keys);
     ScopedSetBool disable_hooks(context().hooks_disabled(),
                                 m_last_insert.disable_hooks);
 
-    push_mode(new InputModes::Insert(*this, m_last_insert.mode, m_last_insert.count));
-    for (auto& key : keys)
-    {
-        // refill last_insert,  this is very inefficient, but necessary at the moment
-        // to properly handle insert completion
-        m_last_insert.keys.push_back(key);
-        current_mode().handle_key(key);
-    }
+    push_mode(new InputModes::Insert(*this, m_last_insert.mode, m_last_insert.count, nullptr));
+    for (auto& key : m_last_insert.keys)
+        handle_key(key);
     kak_assert(dynamic_cast<InputModes::Normal*>(&current_mode()) != nullptr);
 }
 
+void InputHandler::paste(StringView content)
+{
+    current_mode().paste(content);
+}
+
 void InputHandler::prompt(StringView prompt, String initstr, String emptystr,
                           Face prompt_face, PromptFlags flags, char history_register,
                           PromptCompleter completer, PromptCallback callback)
@@ -1645,16 +1600,10 @@ void InputHandler::prompt(StringView prompt, String initstr, String emptystr,
 
 void InputHandler::set_prompt_face(Face prompt_face)
 {
-    InputModes::Prompt* prompt = dynamic_cast<InputModes::Prompt*>(&current_mode());
-    if (prompt)
+    if (auto* prompt = dynamic_cast<InputModes::Prompt*>(&current_mode()))
         prompt->set_prompt_face(prompt_face);
 }
 
-void InputHandler::menu(Vector<DisplayLine> choices, MenuCallback callback)
-{
-    push_mode(new InputModes::Menu(*this, std::move(choices), std::move(callback)));
-}
-
 void InputHandler::on_next_key(StringView mode_name, KeymapMode keymap_mode, KeyCallback callback,
                                Timer::Callback idle_callback)
 {
@@ -1690,8 +1639,7 @@ static bool is_valid(Key key)
 {
     constexpr Key::Modifiers valid_mods = (Key::Modifiers::Control | Key::Modifiers::Alt | Key::Modifiers::Shift);
 
-    return key != Key::Invalid and
-        ((key.modifiers & ~valid_mods) or key.key <= 0x10FFFF);
+    return ((key.modifiers & ~valid_mods) or key.key <= 0x10FFFF);
 }
 
 void InputHandler::handle_key(Key key)
@@ -1701,24 +1649,22 @@ void InputHandler::handle_key(Key key)
 
     const bool was_recording = is_recording();
     ++m_handle_key_level;
-    auto dec = on_scope_end([this]{ --m_handle_key_level; });
+    auto dec = on_scope_end([this]{ --m_handle_key_level;} );
 
-    auto process_key = [&](Key key) {
-        if (m_last_insert.recording)
-            m_last_insert.keys.push_back(key);
-        current_mode().handle_key(key);
-    };
+    if (m_last_insert.recording and m_handle_key_level <= 1)
+        m_last_insert.keys.push_back(key);
 
     const auto keymap_mode = current_mode().keymap_mode();
     KeymapManager& keymaps = m_context.keymaps();
     if (keymaps.is_mapped(key, keymap_mode) and not m_context.keymaps_disabled())
     {
-        ScopedSetBool disable_history{context().history_disabled()};
-        for (auto& k : keymaps.get_mapping(key, keymap_mode).keys)
-            process_key(k);
+        ScopedSetBool noninteractive{context().noninteractive()};
+
+        for (auto& k : keymaps.get_mapping_keys(key, keymap_mode))
+            current_mode().handle_key(k, true);
     }
     else
-        process_key(key);
+        current_mode().handle_key(key, m_handle_key_level > 1);
 
     // do not record the key that made us enter or leave recording mode,
     // and the ones that are triggered recursively by previous keys.
@@ -1733,6 +1679,11 @@ void InputHandler::handle_key(Key key)
     }
 }
 
+void InputHandler::refresh_ifn()
+{
+    current_mode().refresh_ifn();
+}
+
 void InputHandler::start_recording(char reg)
 {
     kak_assert(m_recording_reg == 0);
@@ -1758,9 +1709,9 @@ void InputHandler::stop_recording()
     m_recording_level = -1;
 }
 
-DisplayLine InputHandler::mode_line() const
+ModeInfo InputHandler::mode_info() const
 {
-    return current_mode().mode_line();
+    return current_mode().mode_info();
 }
 
 std::pair<CursorMode, DisplayCoord> InputHandler::get_cursor_info() const
@@ -1788,7 +1739,7 @@ void hide_auto_info_ifn(const Context& context, bool hide)
         context.client().info_hide();
 }
 
-void scroll_window(Context& context, LineCount offset, bool mouse_dragging)
+void scroll_window(Context& context, LineCount offset, OnHiddenCursor on_hidden_cursor)
 {
     Window& window = context.window();
     Buffer& buffer = context.buffer();
@@ -1797,6 +1748,9 @@ void scroll_window(Context& context, LineCount offset, bool mouse_dragging)
     DisplayCoord win_pos = window.position();
     DisplayCoord win_dim = window.dimensions();
 
+    if (on_hidden_cursor == OnHiddenCursor::PreserveSelections)
+        context.ensure_cursor_visible = false;
+
     if ((offset < 0 and win_pos.line == 0) or (offset > 0 and win_pos.line == line_count - 1))
         return;
 
@@ -1806,30 +1760,27 @@ void scroll_window(Context& context, LineCount offset, bool mouse_dragging)
 
     win_pos.line = clamp(win_pos.line + offset, 0_line, line_count-1);
 
-    SelectionList& selections = context.selections();
-    Selection& main_selection = selections.main();
-    const BufferCoord anchor = main_selection.anchor();
-    const BufferCoord cursor = main_selection.cursor();
-
-    auto cursor_off = mouse_dragging ? win_pos.line - window.position().line : 0;
-
-    auto line = clamp(cursor.line + cursor_off, win_pos.line + scrolloff.line,
-                      win_pos.line + win_dim.line - 1 - scrolloff.line);
-    line = clamp(line, 0_line, buffer.line_count() - 1);
-
-    using std::min; using std::max;
-    // This is not exactly a clamp, and must be done in this order as
-    // byte_count_to could return line length
-    auto col = min(max(cursor.column, buffer[line].byte_count_to(win_pos.column)),
-                   buffer[line].length()-1);
-
-    BufferCoord new_cursor = { line, col };
-    BufferCoord new_anchor = (mouse_dragging or new_cursor == cursor) ? anchor : new_cursor;
-
     window.set_position(win_pos);
-    main_selection = { new_anchor, new_cursor };
+    if (on_hidden_cursor != OnHiddenCursor::PreserveSelections)
+    {
+        ScopedSelectionEdition selection_edition{context};
+        SelectionList& selections = context.selections();
+        Selection& main_selection = selections.main();
+        const BufferCoord anchor = main_selection.anchor();
+        const BufferCoordAndTarget cursor = main_selection.cursor();
 
-    selections.sort_and_merge_overlapping();
+        auto cursor_off = win_pos.line - window.position().line;
+
+        auto line = clamp(cursor.line + cursor_off, win_pos.line + scrolloff.line,
+                          win_pos.line + win_dim.line - 1 - scrolloff.line);
+
+        const ColumnCount tabstop = context.options()["tabstop"].get<int>();
+        auto new_cursor = buffer.offset_coord(cursor, line - cursor.line, tabstop);
+
+        main_selection = {on_hidden_cursor == OnHiddenCursor::MoveCursor ? anchor : new_cursor, new_cursor};
+
+        selections.sort_and_merge_overlapping();
+    }
 }
 
 }
