@@ -45,6 +45,9 @@ struct {
     unsigned int version;
     StringView notes;
 } constexpr version_notes[] = { {
+        20240518,
+        "» Fix tests failing on some platforms\n"
+    }, {
         20240509,
         "» {+u}flag-lines -after{} highlighter\n"
         "» asynchronous {+u}shell-script-candidates{} completion\n"
