@@ -44,6 +44,7 @@ constexpr auto enum_desc(Meta::Type<SelectMode>)
         { SelectMode::Append, "append" },
     });
 }
+
 void merge_selections(Selection& sel, const Selection& new_sel)
 {
     const bool forward = sel.cursor() >= sel.anchor();
@@ -74,6 +75,7 @@ UnitTest test_merge_selection{[] {
 template<SelectMode mode, typename T>
 void select(Context& context, T func)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     if (mode == SelectMode::Append)
     {
@@ -140,8 +142,11 @@ void select_and_set_last(Context& context, Func&& func)
 }
 
 template<SelectMode mode = SelectMode::Replace>
-void select_coord(Buffer& buffer, BufferCoord coord, SelectionList& selections)
+void select_coord(Context& context, BufferCoord coord)
 {
+    Buffer& buffer = context.buffer();
+    ScopedSelectionEdition selection_edition{context};
+    SelectionList& selections = context.selections();
     coord = buffer.clamp(coord);
     if (mode == SelectMode::Replace)
         selections = SelectionList{ buffer, coord };
@@ -186,23 +191,31 @@ String build_autoinfo_for_mapping(const Context& context, KeymapMode mode,
     }
 
     for (auto& key : keymaps.get_mapped_keys(mode))
-        descs.emplace_back(to_string(key),
-                           keymaps.get_mapping(key, mode).docstring);
+    {
+        const String& docstring = keymaps.get_mapping_docstring(key, mode);
+        if (keymaps.get_mapping_keys(key, mode).empty() and docstring.empty())
+            continue;
+        if (auto it = find_if(descs, [&](auto& elem) { return elem.second == docstring; });
+            it != descs.end())
+            it->first += ',' + to_string(key);
+        else
+            descs.emplace_back(to_string(key), docstring);
+    }
 
     auto max_len = 0_col;
-    for (auto& desc : descs)
+    for (auto& [keys, docstring] : descs)
     {
-        auto len = desc.first.column_length();
+        auto len = keys.column_length();
         if (len > max_len)
             max_len = len;
     }
 
     String res;
-    for (auto& desc : descs)
+    for (auto& [keys, docstring] : descs)
         res += format("{}:{}{}\n",
-                      desc.first,
-                      String{' ', max_len - desc.first.column_length() + 1},
-                      desc.second);
+                      keys,
+                      String{' ', max_len - keys.column_length() + 1},
+                      docstring);
     return res;
 }
 
@@ -212,7 +225,7 @@ void goto_commands(Context& context, NormalParams params)
     if (params.count != 0)
     {
         context.push_jump();
-        select_coord<mode>(context.buffer(), LineCount{params.count - 1}, context.selections());
+        select_coord<mode>(context, LineCount{params.count - 1});
         if (context.has_window())
             context.window().center_line(LineCount{params.count-1});
     }
@@ -229,7 +242,7 @@ void goto_commands(Context& context, NormalParams params)
             case 'g':
             case 'k':
                 context.push_jump();
-                select_coord<mode>(buffer, BufferCoord{0,0}, context.selections());
+                select_coord<mode>(context, BufferCoord{0,0});
                 break;
             case 'l':
                 select<mode, select_to_line_end<true>>(context, {});
@@ -242,17 +255,17 @@ void goto_commands(Context& context, NormalParams params)
                 break;
             case 'j':
                 context.push_jump();
-                select_coord<mode>(buffer, buffer.line_count() - 1, context.selections());
+                select_coord<mode>(context, buffer.line_count() - 1);
                 break;
             case 'e':
                 context.push_jump();
-                select_coord<mode>(buffer, buffer.back_coord(), context.selections());
+                select_coord<mode>(context, buffer.back_coord());
                 break;
             case 't':
                 if (context.has_window())
                 {
                     auto line = context.window().position().line;
-                    select_coord<mode>(buffer, line, context.selections());
+                    select_coord<mode>(context, line);
                 }
                 break;
             case 'b':
@@ -260,7 +273,7 @@ void goto_commands(Context& context, NormalParams params)
                 {
                     auto& window = context.window();
                     auto line = window.position().line + window.dimensions().line - 1;
-                    select_coord<mode>(buffer, line, context.selections());
+                    select_coord<mode>(context, line);
                 }
                 break;
             case 'c':
@@ -268,7 +281,7 @@ void goto_commands(Context& context, NormalParams params)
                 {
                     auto& window = context.window();
                     auto line = window.position().line + window.dimensions().line / 2;
-                    select_coord<mode>(buffer, line, context.selections());
+                    select_coord<mode>(context, line);
                 }
                 break;
             case 'a':
@@ -328,7 +341,7 @@ void goto_commands(Context& context, NormalParams params)
                     throw runtime_error("no last modification position");
                 if (*pos >= buffer.back_coord())
                     pos = buffer.back_coord();
-                select_coord<mode>(buffer, *pos, context.selections());
+                select_coord<mode>(context, *pos);
                 break;
             }
             default:
@@ -357,6 +370,7 @@ void view_commands(Context& context, NormalParams params)
     const int count = params.count;
     on_next_key_with_autoinfo(context, "view", KeymapMode::View,
                              [count](Key key, Context& context) {
+        context.ensure_cursor_visible = false;
         if (key == Key::Escape)
             return;
 
@@ -389,10 +403,10 @@ void view_commands(Context& context, NormalParams params)
             window.scroll(-std::max<ColumnCount>(1, count));
             break;
         case 'j':
-            scroll_window(context,  std::max<LineCount>(1, count));
+            scroll_window(context,  std::max<LineCount>(1, count), OnHiddenCursor::PreserveSelections);
             break;
         case 'k':
-            scroll_window(context, -std::max<LineCount>(1, count));
+            scroll_window(context, -std::max<LineCount>(1, count), OnHiddenCursor::PreserveSelections);
             break;
         case 'l':
             window.scroll( std::max<ColumnCount>(1, count));
@@ -420,11 +434,12 @@ void replace_with_char(Context& context, NormalParams)
         if (not cp or key == Key::Escape)
             return;
         ScopedEdition edition(context);
+        ScopedSelectionEdition selection_edition{context};
         Buffer& buffer = context.buffer();
         context.selections().for_each([&](size_t index, Selection& sel) {
             CharCount count = char_length(buffer, sel);
             replace(buffer, sel, String{*cp, count});
-        });
+        }, false);
     }, "replace with char", "enter char to replace with\n");
 }
 
@@ -440,6 +455,7 @@ void for_each_codepoint(Context& context, NormalParams)
     using Utf8It = utf8::iterator<BufferIterator>;
 
     ScopedEdition edition(context);
+    ScopedSelectionEdition selection_edition{context};
     Buffer& buffer = context.buffer();
 
     context.selections().for_each([&](size_t index, Selection& sel) {
@@ -449,7 +465,7 @@ void for_each_codepoint(Context& context, NormalParams)
              begin != end; ++begin)
             utf8::dump(std::back_inserter(str), func(*begin));
         replace(buffer, sel, str);
-    });
+    }, false);
 }
 
 void command(const Context& context, EnvVarMap env_vars, char reg = 0)
@@ -457,17 +473,15 @@ void command(const Context& context, EnvVarMap env_vars, char reg = 0)
     if (not CommandManager::has_instance())
         throw runtime_error{"commands are not supported"};
 
-    CommandManager::instance().clear_last_complete_command();
-
     String default_command = context.main_sel_register_value(reg ? reg : ':').str();
 
     context.input_handler().prompt(
         ":", {}, default_command,
         context.faces()["Prompt"], PromptFlags::DropHistoryEntriesWithBlankPrefix,
         ':',
-        [](const Context& context, CompletionFlags flags,
-           StringView cmd_line, ByteCount pos) {
-               return CommandManager::instance().complete(context, flags, cmd_line, pos);
+        [completer=CommandManager::Completer{}](const Context& context, CompletionFlags flags,
+           StringView cmd_line, ByteCount pos) mutable {
+               return completer(context, flags, cmd_line, pos);
         },
         [env_vars = std::move(env_vars), default_command](StringView cmdline, PromptEvent event, Context& context) {
             if (context.has_client())
@@ -553,7 +567,8 @@ void pipe(Context& context, NormalParams params)
         prompt, {}, default_command, context.faces()["Prompt"],
         PromptFlags::DropHistoryEntriesWithBlankPrefix, '|',
         shell_complete,
-        [default_command](StringView cmdline, PromptEvent event, Context& context)
+        [default_command, selection_edition=std::make_shared<ScopedSelectionEdition>(context)]
+        (StringView cmdline, PromptEvent event, Context& context)
         {
             if (event != PromptEvent::Validate)
                 return;
@@ -635,6 +650,7 @@ void erase_selections(Context& context, NormalParams params)
         RegisterManager::instance()[reg].set(context, context.selections_content());
     }
     ScopedEdition edition(context);
+    ScopedSelectionEdition selection_edition{context};
     context.selections().erase();
 }
 
@@ -649,21 +665,16 @@ void change(Context& context, NormalParams params)
     enter_insert_mode<InsertMode::Replace>(context, params);
 }
 
-enum class PasteMode
-{
-    Append,
-    Insert,
-    Replace
-};
-
-BufferCoord paste_pos(Buffer& buffer, const Selection& sel, PasteMode mode, bool linewise)
+BufferCoord paste_pos(Buffer& buffer, BufferCoord min, BufferCoord max, PasteMode mode, bool linewise)
 {
     switch (mode)
     {
         case PasteMode::Append:
-            return linewise ? std::min(buffer.line_count(), sel.max().line+1) : buffer.char_next(sel.max());
+            if (buffer.is_end(max))
+                return max;
+            return linewise ? std::min(buffer.line_count(), max.line+1) : buffer.char_next(max);
         case PasteMode::Insert:
-            return linewise ? sel.min().line : sel.min();
+            return linewise ? min.line : min;
         default:
             kak_assert(false);
             return {};
@@ -681,16 +692,18 @@ void paste(Context& context, NormalParams params)
 
     auto& buffer = context.buffer();
     ScopedEdition edition(context);
-    context.selections().for_each([&](size_t index, Selection& sel) {
-        auto& str = strings[std::min(strings.size()-1, index)];
+    ScopedSelectionEdition selection_edition{context};
+    context.selections().for_each([&, last=BufferCoord{}](size_t index, Selection& sel) mutable {
+        auto& str = strings[index % strings.size()];
         auto& min = sel.min();
         auto& max = sel.max();
         BufferRange range = (mode == PasteMode::Replace) ?
             buffer.replace(min, buffer.char_next(max), str)
-          : buffer.insert(paste_pos(buffer, sel, mode, linewise), str);
+          : buffer.insert(paste_pos(buffer, min, std::max(max, last), mode, linewise), str);
         min = range.begin;
         max = range.end > range.begin ? buffer.char_prev(range.end) : range.begin;
-    });
+        last = max;
+    }, mode == PasteMode::Append);
 }
 
 template<PasteMode mode>
@@ -717,13 +730,14 @@ void paste_all(Context& context, NormalParams params)
 
     Buffer& buffer = context.buffer();
     Vector<Selection> result;
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     {
         ScopedEdition edition(context);
         selections.for_each([&](size_t, const Selection& sel) {
             auto range = (mode == PasteMode::Replace) ?
                 buffer.replace(sel.min(), buffer.char_next(sel.max()), all)
-              : buffer.insert(paste_pos(buffer, sel, mode, linewise), all);
+              : buffer.insert(paste_pos(buffer, sel.min(), sel.max(), mode, linewise), all);
 
             ByteCount pos_offset = 0;
             BufferCoord pos = range.begin;
@@ -734,7 +748,7 @@ void paste_all(Context& context, NormalParams params)
                 pos = buffer.next(end);
                 pos_offset = offset;
             }
-        });
+        }, mode == PasteMode::Append);
     }
     selections = std::move(result);
 }
@@ -749,7 +763,8 @@ void insert_output(Context& context, NormalParams params)
         prompt, {}, default_command, context.faces()["Prompt"],
         PromptFlags::DropHistoryEntriesWithBlankPrefix, '|',
         shell_complete,
-        [default_command](StringView cmdline, PromptEvent event, Context& context)
+        [default_command, selection_edition=std::make_shared<ScopedSelectionEdition>(context)]
+        (StringView cmdline, PromptEvent event, Context& context)
         {
             if (event != PromptEvent::Validate)
                 return;
@@ -773,10 +788,10 @@ void insert_output(Context& context, NormalParams params)
 
                 auto& min = sel.min();
                 auto& max = sel.max();
-                auto range = insert(buffer, sel, paste_pos(buffer, sel, mode, false), out);
+                auto range = insert(buffer, sel, paste_pos(buffer, min, max, mode, false), out);
                 min = range.begin;
                 max = range.end > range.begin ? buffer.char_prev(range.end) : range.begin;
-            });
+            }, mode == PasteMode::Append);
             selections.set_main_index(old_main);
         });
 }
@@ -821,7 +836,9 @@ void regex_prompt(Context& context, String prompt, char reg, T func)
                        [&](auto&& m) { candidates.push_back(m.candidate().str()); return true; });
             return {(int)(word.begin() - regex.begin()), pos,  std::move(candidates) };
         },
-        [=, func=T(std::move(func))](StringView str, PromptEvent event, Context& context) mutable {
+        [=, func=T(std::move(func)), saved_reg=RegisterManager::instance()[reg].save(context),
+         selection_edition=std::make_shared<ScopedSelectionEdition>(context)]
+        (StringView str, PromptEvent event, Context& context) mutable {
             try
             {
                 if (event != PromptEvent::Change and context.has_client())
@@ -836,16 +853,25 @@ void regex_prompt(Context& context, String prompt, char reg, T func)
                         context.window().set_position(position);
 
                     context.input_handler().set_prompt_face(context.faces()["Prompt"]);
+                    RegisterManager::instance()[reg].restore(context, saved_reg);
                 }
 
-                if (not incsearch and event == PromptEvent::Change)
-                    return;
-
-                if (event == PromptEvent::Validate)
+                switch (event)
+                {
+                case PromptEvent::Abort: return;
+                case PromptEvent::Change:
+                    if (not incsearch)
+                        return;
+                    if (not str.empty())
+                        RegisterManager::instance()[reg].set(context, str.str());
+                    break;
+                case PromptEvent::Validate:
+                    if (not str.empty())
+                        RegisterManager::instance()[reg].set(context, str.str());
                     context.push_jump();
-
-                if (not str.empty() or event == PromptEvent::Validate)
-                    func(Regex{str.empty() ? default_regex : str, direction_flags(mode)}, event, context);
+                    break;
+                }
+                func(Regex{str.empty() ? default_regex : str, direction_flags(mode)}, event, context);
             }
             catch (regex_error& err)
             {
@@ -916,13 +942,8 @@ void search(Context& context, NormalParams params)
     const int count = params.count;
 
     regex_prompt<regex_mode>(context, prompt.str(), reg,
-                 [reg, count, saved_reg = RegisterManager::instance()[reg].save(context)]
+                 [count]
                  (const Regex& regex, PromptEvent event, Context& context) {
-                     RegisterManager::instance()[reg].restore(context, saved_reg);
-                     if (event == PromptEvent::Abort)
-                         return;
-                     RegisterManager::instance()[reg].set(context, regex.str());
-
                      if (regex.empty() or regex.str().empty())
                          return;
 
@@ -941,6 +962,7 @@ void search_next(Context& context, NormalParams params)
     if (not str.empty())
     {
         Regex regex{str, direction_flags(regex_mode)};
+        ScopedSelectionEdition selection_edition{context};
         auto& selections = context.selections();
         bool main_wrapped = false;
         do {
@@ -979,7 +1001,7 @@ void use_selection_as_search_pattern(Context& context, NormalParams params)
                       smart and is_bow(buffer, beg) ? "\\b" : "",
                       escape(buffer.string(beg, end), "^$\\.*+?()[]{}|", '\\'),
                       smart and is_eow(buffer, end) ? "\\b" : "");
-    }) | gather<HashSet>();
+    }) | gather<HashSet<String>>();
     String pattern = join(patterns, '|', false);
 
     const char reg = to_lower(params.reg ? params.reg : '/');
@@ -991,8 +1013,8 @@ void use_selection_as_search_pattern(Context& context, NormalParams params)
     RegisterManager::instance()[reg].set(context, {pattern});
 
     // Hack, as Window do not take register state into account
-    if (context.has_window())
-        context.window().force_redraw();
+    if (context.has_client())
+        context.client().force_redraw();
 }
 
 void select_regex(Context& context, NormalParams params)
@@ -1002,14 +1024,8 @@ void select_regex(Context& context, NormalParams params)
     auto prompt = capture ? format("select (capture {}):", capture) :  "select:"_str;
 
     regex_prompt(context, std::move(prompt), reg,
-                 [reg, capture, saved_reg = RegisterManager::instance()[reg].save(context)]
+                 [capture]
                  (Regex ex, PromptEvent event, Context& context) {
-        RegisterManager::instance()[reg].restore(context, saved_reg);
-        if (event == PromptEvent::Abort)
-            return;
-        if (not context.history_disabled())
-            RegisterManager::instance()[reg].set(context, ex.str());
-
         auto& selections = context.selections();
         auto& buffer = selections.buffer();
         if (not ex.empty() and not ex.str().empty())
@@ -1024,14 +1040,8 @@ void split_regex(Context& context, NormalParams params)
     auto prompt = capture ? format("split (on capture {}):", (int)capture) :  "split:"_str;
 
     regex_prompt(context, std::move(prompt), reg,
-                 [reg, capture, saved_reg = RegisterManager::instance()[reg].save(context)]
+                 [capture]
                  (Regex ex, PromptEvent event, Context& context) {
-        RegisterManager::instance()[reg].restore(context, saved_reg);
-        if (event == PromptEvent::Abort)
-            return;
-        if (not context.history_disabled())
-            RegisterManager::instance()[reg].set(context, ex.str());
-
         auto& selections = context.selections();
         auto& buffer = selections.buffer();
         if (not ex.empty() and not ex.str().empty())
@@ -1042,6 +1052,7 @@ void split_regex(Context& context, NormalParams params)
 void split_lines(Context& context, NormalParams params)
 {
     const LineCount count{params.count == 0 ? 1 : params.count};
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     auto& buffer = context.buffer();
     Vector<Selection> res;
@@ -1068,6 +1079,7 @@ void split_lines(Context& context, NormalParams params)
 
 void select_boundaries(Context& context, NormalParams)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     Vector<Selection> res;
     for (auto& sel : selections)
@@ -1083,6 +1095,7 @@ void join_lines_select_spaces(Context& context, NormalParams)
 {
     auto& buffer = context.buffer();
     Vector<Selection> selections;
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
     {
         const LineCount min_line = sel.min().line;
@@ -1106,6 +1119,7 @@ void join_lines_select_spaces(Context& context, NormalParams)
 
 void join_lines(Context& context, NormalParams params)
 {
+    ScopedSelectionEdition selection_edition{context};
     SelectionList sels{context.selections()};
     auto restore_sels = on_scope_end([&]{
         sels.update();
@@ -1123,14 +1137,8 @@ void keep(Context& context, NormalParams params)
     const char reg = to_lower(params.reg ? params.reg : '/');
 
     regex_prompt(context, prompt.str(), reg,
-                 [reg, saved_reg = RegisterManager::instance()[reg].save(context)]
+                 [selection_edition=std::make_shared<ScopedSelectionEdition>(context)]
                  (const Regex& regex, PromptEvent event, Context& context) {
-        RegisterManager::instance()[reg].restore(context, saved_reg);
-        if (event == PromptEvent::Abort)
-            return;
-        if (not context.history_disabled())
-            RegisterManager::instance()[reg].set(context, regex.str());
-
         if (regex.empty() or regex.str().empty())
             return;
 
@@ -1145,7 +1153,7 @@ void keep(Context& context, NormalParams params)
             const auto flags = match_flags(is_bol(begin.coord()), false,
                                            is_bow(buffer, begin.coord()),
                                            is_eow(buffer, end.coord()));
-            if (regex_search(begin, end, begin, end, regex, flags) == matching)
+            if (regex_search(begin, end, begin, end, regex, flags, EventManager::handle_urgent_events) == matching)
                 keep.push_back(sel);
         }
         if (keep.empty())
@@ -1160,7 +1168,8 @@ void keep_pipe(Context& context, NormalParams params)
     context.input_handler().prompt(
         "keep pipe:", {}, default_command, context.faces()["Prompt"],
         PromptFlags::DropHistoryEntriesWithBlankPrefix, '|', shell_complete,
-        [default_command](StringView cmdline, PromptEvent event, Context& context) {
+        [default_command, selection_edition=std::make_shared<ScopedSelectionEdition>(context)]
+        (StringView cmdline, PromptEvent event, Context& context) {
             if (event != PromptEvent::Validate)
                 return;
 
@@ -1206,6 +1215,7 @@ void indent(Context& context, NormalParams params)
     ScopedEdition edition(context);
     auto& buffer = context.buffer();
     LineCount last_line = 0;
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
     {
         for (auto line = std::max(last_line, sel.min().line); line < sel.max().line+1; ++line)
@@ -1230,6 +1240,7 @@ void deindent(Context& context, NormalParams params)
 
     auto& buffer = context.buffer();
     LineCount last_line = 0;
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
     {
         for (auto line = std::max(sel.min().line, last_line);
@@ -1395,8 +1406,6 @@ void select_object(Context& context, NormalParams params)
          {{alt(';')},    "run command in object context"}}));
 }
 
-enum Direction { Backward = -1, Forward = 1 };
-
 template<Direction direction, bool half = false>
 void scroll(Context& context, NormalParams params)
 {
@@ -1404,12 +1413,13 @@ void scroll(Context& context, NormalParams params)
     const int count = params.count ? params.count : 1;
     const LineCount offset = (window.dimensions().line - 2) / (half ? 2 : 1) * count;
 
-    scroll_window(context, offset * direction);
+    scroll_window(context, offset * direction, OnHiddenCursor::MoveCursorAndAnchor);
 }
 
 template<Direction direction>
 void copy_selections_on_next_lines(Context& context, NormalParams params)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     auto& buffer = context.buffer();
     const ColumnCount tabstop = context.options()["tabstop"].get<int>();
@@ -1463,6 +1473,7 @@ template<Direction direction>
 void rotate_selections(Context& context, NormalParams params)
 {
     const int count = params.count ? params.count : 1;
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     const int index = selections.main_index();
     const int num = selections.size();
@@ -1480,6 +1491,7 @@ void rotate_selections_content(Context& context, NormalParams params)
     if (group == 0 or group > (int)strings.size())
         group = (int)strings.size();
     count = count % group;
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     auto main = strings.begin() + selections.main_index();
     for (auto it = strings.begin(); it != strings.end(); )
@@ -1564,6 +1576,7 @@ void replay_macro(Context& context, NormalParams params)
 
     auto keys = parse_keys(reg_val[0]);
     ScopedEdition edition(context);
+    ScopedSelectionEdition selection_edition{context};
     do
     {
         for (auto& key : keys)
@@ -1581,6 +1594,7 @@ void jump(Context& context, NormalParams params)
 
     Buffer* oldbuf = &context.buffer();
     Buffer& buffer = const_cast<Buffer&>(jump.buffer());
+    ScopedSelectionEdition selection_edition{context};
     if (&buffer != oldbuf)
         context.change_buffer(buffer);
     context.selections_write_only() = jump;
@@ -1595,6 +1609,7 @@ void push_selections(Context& context, NormalParams)
 
 void align(Context& context, NormalParams)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     auto& buffer = context.buffer();
     const ColumnCount tabstop = context.options()["tabstop"].get<int>();
@@ -1648,6 +1663,7 @@ void copy_indent(Context& context, NormalParams params)
 {
     int selection = params.count;
     auto& buffer = context.buffer();
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     Vector<LineCount> lines;
     for (auto sel : selections)
@@ -1688,6 +1704,8 @@ void tabs_to_spaces(Context& context, NormalParams params)
     const ColumnCount tabstop = params.count == 0 ? opt_tabstop : params.count;
     Vector<Selection> tabs;
     Vector<String> spaces;
+    ScopedSelectionEdition selection_edition{context};
+    ScopedEdition edition{context};
     for (auto& sel : context.selections())
     {
         for (auto it = buffer.iterator_at(sel.min()),
@@ -1712,6 +1730,8 @@ void spaces_to_tabs(Context& context, NormalParams params)
     const ColumnCount opt_tabstop = context.options()["tabstop"].get<int>();
     const ColumnCount tabstop = params.count == 0 ? opt_tabstop : params.count;
     Vector<Selection> spaces;
+    ScopedSelectionEdition selection_edition{context};
+    ScopedEdition edition{context};
     for (auto& sel : context.selections())
     {
         for (auto it = buffer.iterator_at(sel.min()),
@@ -1744,15 +1764,17 @@ void spaces_to_tabs(Context& context, NormalParams params)
 
 void trim_selections(Context& context, NormalParams)
 {
+    using Utf8It = utf8::iterator<BufferIterator>;
     auto& buffer = context.buffer();
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     Vector<int> to_remove;
 
     for (int i = 0; i < (int)selections.size(); ++i)
     {
         auto& sel = selections[i];
-        auto beg = buffer.iterator_at(sel.min());
-        auto end = buffer.iterator_at(sel.max());
+        auto beg = Utf8It{buffer.iterator_at(sel.min()), buffer};
+        auto end = Utf8It{buffer.iterator_at(sel.max()), buffer};
         while (beg != end and is_blank(*beg))
             ++beg;
         while (beg != end and is_blank(*end))
@@ -1762,8 +1784,8 @@ void trim_selections(Context& context, NormalParams)
             to_remove.push_back(i);
         else
         {
-            sel.min() = beg.coord();
-            sel.max() = end.coord();
+            sel.min() = beg.base().coord();
+            sel.max() = end.base().coord();
         }
     }
 
@@ -1773,7 +1795,7 @@ void trim_selections(Context& context, NormalParams)
         selections.remove(i);
 }
 
-SelectionList read_selections_from_register(char reg, Context& context)
+SelectionList read_selections_from_register(char reg, const Context& context)
 {
     if (not is_basic_alpha(reg) and reg != '^')
         throw runtime_error("selections can only be saved to the '^' and alphabetic registers");
@@ -1868,6 +1890,7 @@ void combine_selections(Context& context, SelectionList list, Func func, StringV
                                      return;
 
                                  const auto op = key_to_combine_op(key);
+                                 ScopedSelectionEdition selection_edition{context};
                                  auto& sels = context.selections();
                                  list.update();
                                  if (op == CombineOp::Append)
@@ -1921,7 +1944,10 @@ void save_selections(Context& context, NormalParams params)
     };
 
     if (combine and not empty)
+    {
+        ScopedSelectionEdition selection_edition{context};
         combine_selections(context, read_selections_from_register(reg, context), save_to_reg, "combine selections to register");
+    }
     else
         save_to_reg(context, context.selections());
 }
@@ -1932,6 +1958,8 @@ void restore_selections(Context& context, NormalParams params)
     const char reg = to_lower(params.reg ? params.reg : '^');
     auto selections = read_selections_from_register(reg, context);
 
+    ScopedSelectionEdition selection_edition{context};
+
     auto set_selections = [reg](Context& context, SelectionList sels) {
         auto size = sels.size();
         context.selections_write_only() = std::move(sels);
@@ -2001,6 +2029,14 @@ void move_in_history(Context& context, NormalParams params)
                             history_id, max_history_id));
 }
 
+template<Direction direction>
+void undo_selection_change(Context& context, NormalParams params)
+{
+    int count = std::max(1, params.count);
+    while (count--)
+        context.undo_selection_change<direction>();
+}
+
 void exec_user_mappings(Context& context, NormalParams params)
 {
     on_next_key_with_autoinfo(context, "user-mapping", KeymapMode::None,
@@ -2008,14 +2044,14 @@ void exec_user_mappings(Context& context, NormalParams params)
         if (not context.keymaps().is_mapped(key, KeymapMode::User))
             return;
 
-        auto& mapping = context.keymaps().get_mapping(key, KeymapMode::User);
         ScopedSetBool disable_keymaps(context.keymaps_disabled());
-        ScopedSetBool disable_history(context.history_disabled());
+        ScopedSetBool noninteractive(context.noninteractive());
 
         InputHandler::ScopedForceNormal force_normal{context.input_handler(), params};
 
         ScopedEdition edition(context);
-        for (auto& key : mapping.keys)
+        ScopedSelectionEdition selection_edition{context};
+        for (auto& key : context.keymaps().get_mapping_keys(key, KeymapMode::User))
             context.input_handler().handle_key(key);
     }, "user mapping",
     build_autoinfo_for_mapping(context, KeymapMode::User, {}));
@@ -2027,6 +2063,7 @@ void add_empty_line(Context& context, NormalParams params)
     int count = std::max(params.count, 1);
     String new_lines{'\n', CharCount{count}};
     auto& buffer = context.buffer();
+    ScopedSelectionEdition selection_edition{context};
     auto& sels = context.selections();
     ScopedEdition edition{context};
     for (int i = 0; i < sels.size(); ++i)
@@ -2045,6 +2082,7 @@ public:
     void operator() (Context& context, NormalParams params)
     {
         ScopedEdition edition(context);
+        ScopedSelectionEdition selection_edition{context};
         do { m_func(context, {0, params.reg}); } while(--params.count > 0);
     }
 private:
@@ -2055,6 +2093,7 @@ template<void (*func)(Context&, NormalParams)>
 void repeated(Context& context, NormalParams params)
 {
     ScopedEdition edition(context);
+    ScopedSelectionEdition selection_edition{context};
     do { func(context, {0, params.reg}); } while(--params.count > 0);
 }
 
@@ -2064,6 +2103,7 @@ void move_cursor(Context& context, NormalParams params)
     kak_assert(mode == SelectMode::Replace or mode == SelectMode::Extend);
     const Type offset{direction * std::max(params.count,1)};
     const ColumnCount tabstop = context.options()["tabstop"].get<int>();
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     for (auto& sel : selections)
     {
@@ -2077,11 +2117,13 @@ void move_cursor(Context& context, NormalParams params)
 void select_whole_buffer(Context& context, NormalParams)
 {
     auto& buffer = context.buffer();
+    ScopedSelectionEdition selection_edition{context};
     context.selections_write_only() = SelectionList{buffer, {{0,0}, {buffer.back_coord(), max_column}}};
 }
 
 void keep_selection(Context& context, NormalParams p)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     const int index = p.count ? p.count-1 : selections.main_index();
     if (index >= selections.size())
@@ -2093,6 +2135,7 @@ void keep_selection(Context& context, NormalParams p)
 
 void remove_selection(Context& context, NormalParams p)
 {
+    ScopedSelectionEdition selection_edition{context};
     auto& selections = context.selections();
     const int index = p.count ? p.count-1 : selections.main_index();
     if (index >= selections.size())
@@ -2106,12 +2149,14 @@ void remove_selection(Context& context, NormalParams p)
 
 void clear_selections(Context& context, NormalParams)
 {
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
         sel.anchor() = sel.cursor();
 }
 
 void flip_selections(Context& context, NormalParams)
 {
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
     {
         const BufferCoord tmp = sel.anchor();
@@ -2123,6 +2168,7 @@ void flip_selections(Context& context, NormalParams)
 
 void ensure_forward(Context& context, NormalParams)
 {
+    ScopedSelectionEdition selection_edition{context};
     for (auto& sel : context.selections())
     {
         const BufferCoord min = sel.min(), max = sel.max();
@@ -2134,31 +2180,42 @@ void ensure_forward(Context& context, NormalParams)
 
 void merge_consecutive(Context& context, NormalParams params)
 {
+    ScopedSelectionEdition selection_edition{context};
     ensure_forward(context, params);
     context.selections().merge_consecutive();
 }
 
 void merge_overlapping(Context& context, NormalParams params)
 {
+    ScopedSelectionEdition selection_edition{context};
     ensure_forward(context, params);
     context.selections().merge_overlapping();
 }
 
 void duplicate_selections(Context& context, NormalParams params)
 {
+    ScopedSelectionEdition selection_edition{context};
     SelectionList& sels = context.selections();
     Vector<Selection> new_sels;
     const int count = params.count ? params.count : 2;
+    BasicSelection last{BufferCoord{-1,-1}};
+    size_t index = 0;
+    size_t main_index = 0;
     for (const auto& sel : sels)
-        new_sels.insert(new_sels.end(), count, sel);
-    context.selections().set(std::move(new_sels), sels.main_index() * count);
+    {
+        new_sels.insert(new_sels.end(), sel != last ? count : 1, sel);
+        last = sel;
+        if (index++ == sels.main_index())
+            main_index = new_sels.size() - 1;
+    }
+    context.selections().set(std::move(new_sels), main_index);
 }
 
 void force_redraw(Context& context, NormalParams)
 {
     if (context.has_client())
     {
-        context.client().force_redraw();
+        context.client().force_redraw(true);
         context.client().redraw_ifn();
     }
 }
@@ -2309,8 +2366,11 @@ static constexpr HashMap<Key, NormalCmd, MemoryDomain::Undefined, KeymapBackend>
 
     { {'u'}, {"undo", undo} },
     { {'U'}, {"redo", redo} },
-    { {alt('u')}, {"move backward in history", move_in_history<Direction::Backward>} },
-    { {alt('U')}, {"move forward in history", move_in_history<Direction::Forward>} },
+    { {ctrl('k')}, {"move backward in history", move_in_history<Direction::Backward>} },
+    { {ctrl('j')}, {"move forward in history", move_in_history<Direction::Forward>} },
+
+    { {alt('u')}, {"undo selection change", undo_selection_change<Backward>} },
+    { {alt('U')}, {"redo selection change", undo_selection_change<Forward>} },
 
     { {alt('i')}, {"select inner object", select_object<ObjectFlags::ToBegin | ObjectFlags::ToEnd | ObjectFlags::Inner>} },
     { {alt('a')}, {"select whole object", select_object<ObjectFlags::ToBegin | ObjectFlags::ToEnd>} },
