@@ -2,6 +2,7 @@
 
 #include "buffer_utils.hh"
 #include "context.hh"
+#include "event_manager.hh"
 #include "flags.hh"
 #include "option_types.hh"
 #include "regex.hh"
@@ -287,16 +288,19 @@ find_opening(Iterator pos, const Container& container,
     // When on the token of a non-nestable block, we want to consider it opening
     if (nestable and
         backward_regex_search(container.begin(), pos,
-                              container.begin(), container.end(), res, closing) and
+                              container.begin(), container.end(), res, closing,
+                              RegexExecFlags::None, EventManager::handle_urgent_events) and
         res[0].second == pos)
         pos = res[0].first;
 
-    using RegexIt = RegexIterator<Iterator, RegexMode::Backward>;
-    for (auto&& match : RegexIt{container.begin(), pos, container.begin(), container.end(), opening})
+    using RegexIt = RegexIterator<Iterator, RegexMode::Backward, const Regex, void (*)()>;
+    for (auto&& match : RegexIt{container.begin(), pos, container.begin(), container.end(), opening,
+                                RegexExecFlags::None, EventManager::handle_urgent_events})
     {
         if (nestable)
         {
-            for (auto m [[maybe_unused]] : RegexIt{match[0].second, pos, container.begin(), container.end(), closing})
+            for (auto m [[maybe_unused]] : RegexIt{match[0].second, pos, container.begin(), container.end(), closing,
+                                                   RegexExecFlags::None, EventManager::handle_urgent_events})
                 ++level;
         }
 
@@ -314,12 +318,13 @@ find_closing(Iterator pos, const Container& container,
              const Regex& opening, const Regex& closing,
              int level, bool nestable)
 {
-    using RegexIt = RegexIterator<Iterator, RegexMode::Forward>;
-    for (auto match : RegexIt{pos, container.end(), container.begin(), container.end(), closing})
+    for (auto match : RegexIterator{pos, container.end(), container.begin(), container.end(), closing,
+                                    RegexExecFlags::None, EventManager::handle_urgent_events})
     {
         if (nestable)
         {
-            for (auto m [[maybe_unused]] : RegexIt{pos, match[0].first, container.begin(), container.end(), opening})
+            for (auto m [[maybe_unused]] : RegexIterator{pos, match[0].first, container.begin(), container.end(), opening,
+                                                         RegexExecFlags::None, EventManager::handle_urgent_events})
                 ++level;
         }
 
@@ -356,7 +361,8 @@ find_surrounding(const Container& container, Iterator pos,
             return {};
     }
     else if (MatchResults<Iterator> res;
-             regex_search(pos, container.end(), container.begin(), container.end(), res, opening) and
+             regex_search(pos, container.end(), container.begin(), container.end(), res, opening,
+                          RegexExecFlags::None, EventManager::handle_urgent_events) and
              res[0].first == pos) // Skip opening match if pos lies on it
         last = empty(res[0]) ? std::next(res[0].second) : res[0].second;
 
@@ -420,6 +426,12 @@ select_to_reverse(const Context& context, const Selection& selection,
                   Codepoint c, int count, bool inclusive)
 {
     auto& buffer = context.buffer();
+
+    // if we are selecting backwards from the beginning of the buffer,
+    // there is nothing more that can be selected.
+    if (selection.cursor() == buffer.begin())
+        return {};
+
     Utf8Iterator begin{buffer.iterator_at(selection.cursor()), buffer};
     Utf8Iterator end = begin;
     do
@@ -483,7 +495,7 @@ select_sentence(const Context& context, const Selection& selection,
     BufferIterator first = buffer.iterator_at(selection.cursor());
     BufferIterator last;
 
-    for (++count; count > 0; --count)
+    for (int i = 0; i <= count; ++i)
     {
         if (not (flags & ObjectFlags::ToEnd) and first != buffer.begin())
         {
@@ -494,7 +506,7 @@ select_sentence(const Context& context, const Selection& selection,
                 first = prev_non_blank;
         }
 
-        if (last == BufferIterator{})
+        if (i == 0)
             last = first;
 
         if (flags & ObjectFlags::ToBegin)
@@ -552,7 +564,7 @@ select_paragraph(const Context& context, const Selection& selection,
     BufferIterator first = buffer.iterator_at(selection.cursor());
     BufferIterator last;
 
-    for (++count; count > 0; --count)
+    for (int i = 0; i <= count; ++i)
     {
         if (not (flags & ObjectFlags::ToEnd) and first.coord() > BufferCoord{0,1} and
             is_eol(*(first-1)) and first-1 != buffer.begin() and is_eol(*(first-2)))
@@ -561,7 +573,7 @@ select_paragraph(const Context& context, const Selection& selection,
                  first != buffer.begin() and (first+1) != buffer.end() and
                  is_eol(*(first-1)) and is_eol(*first))
             ++first;
-        if (last == BufferIterator{})
+        if (i == 0)
             last = first;
 
         if ((flags & ObjectFlags::ToBegin) and first != buffer.begin())
@@ -862,11 +874,13 @@ static bool find_next(const Buffer& buffer, const BufferIterator& pos,
 {
     if (pos != buffer.end() and
         regex_search(pos, buffer.end(), buffer.begin(), buffer.end(),
-                     matches, ex, match_flags(buffer, pos, buffer.end())))
+                     matches, ex, match_flags(buffer, pos, buffer.end()),
+                     EventManager::handle_urgent_events))
         return true;
     wrapped = true;
     return regex_search(buffer.begin(), buffer.end(), buffer.begin(), buffer.end(),
-                        matches, ex, match_flags(buffer, buffer.begin(), buffer.end()));
+                        matches, ex, match_flags(buffer, buffer.begin(), buffer.end()),
+                        EventManager::handle_urgent_events);
 }
 
 static bool find_prev(const Buffer& buffer, const BufferIterator& pos,
@@ -877,13 +891,15 @@ static bool find_prev(const Buffer& buffer, const BufferIterator& pos,
         backward_regex_search(buffer.begin(), pos, buffer.begin(), buffer.end(),
                               matches, ex,
                               match_flags(buffer, buffer.begin(), pos) |
-                              RegexExecFlags::NotInitialNull))
+                              RegexExecFlags::NotInitialNull,
+                              EventManager::handle_urgent_events))
         return true;
     wrapped = true;
     return backward_regex_search(buffer.begin(), buffer.end(), buffer.begin(), buffer.end(),
                                  matches, ex,
                                  match_flags(buffer, buffer.begin(), buffer.end()) |
-                                 RegexExecFlags::NotInitialNull);
+                                 RegexExecFlags::NotInitialNull,
+                                 EventManager::handle_urgent_events);
 }
 
 template<RegexMode mode>
@@ -929,7 +945,8 @@ Vector<Selection> select_matches(const Buffer& buffer, ConstArrayView<Selection>
         auto sel_beg = buffer.iterator_at(sel.min());
         auto sel_end = utf8::next(buffer.iterator_at(sel.max()), buffer.end());
 
-        for (auto&& match : RegexIterator{sel_beg, sel_end, vm, match_flags(buffer, sel_beg, sel_end)})
+        for (auto&& match : RegexIterator{sel_beg, sel_end, vm, match_flags(buffer, sel_beg, sel_end),
+                                          EventManager::handle_urgent_events})
         {
             auto capture = match[capture_idx];
             if (not capture.matched or capture.first == sel_end)
@@ -965,7 +982,8 @@ Vector<Selection> split_on_matches(const Buffer& buffer, ConstArrayView<Selectio
     ThreadedRegexVM<BufferIterator, RegexMode::Forward | RegexMode::Search> vm{*regex.impl()};
     for (auto& sel : selections)
     {
-        auto begin = buffer.iterator_at(sel.min());
+        auto sel_begin = buffer.iterator_at(sel.min());
+        auto begin = sel_begin;
         auto sel_end = utf8::next(buffer.iterator_at(sel.max()), buf_end);
 
         for (auto&& match : RegexIterator{begin, sel_end, vm, match_flags(buffer, begin, sel_end)})
@@ -975,7 +993,7 @@ Vector<Selection> split_on_matches(const Buffer& buffer, ConstArrayView<Selectio
             if (end == buf_end)
                 continue;
 
-            if (end != buf_begin)
+            if (end != sel_begin)
             {
                 auto sel_end = (begin == end) ? end : utf8::previous(end, begin);
                 result.push_back(keep_direction({ begin.coord(), sel_end.coord() }, sel));
