@@ -1,38 +1,25 @@
 #include "shared_string.hh"
 #include "buffer_utils.hh"
 
-#include <cstring>
-
 namespace Kakoune
 {
 
-StringDataPtr StringData::create(ArrayView<const StringView> strs)
+StringDataPtr StringData::Registry::intern(StringView str, size_t hash)
 {
-    const int len = accumulate(strs, 0, [](int l, StringView s) {
-                        return l + (int)s.length();
-                    });
-    void* ptr = StringData::operator new(sizeof(StringData) + len + 1);
-    auto* res = new (ptr) StringData(len);
-    auto* data = reinterpret_cast<char*>(res + 1);
-    for (auto& str : strs)
-    {
-        memcpy(data, str.begin(), (size_t)str.length());
-        data += (int)str.length();
-    }
-    *data = 0;
-    return RefPtr<StringData, PtrPolicy>{res};
+    kak_assert(hash_value(str) == hash);
+    auto index = m_strings.find_index(str, hash);
+    if (index >= 0)
+        return StringDataPtr{m_strings.item(index).value};
+
+    auto data = StringData::create(str);
+    data->refcount |= interned_flag;
+    m_strings.insert({data->strview(), data.get()}, hash);
+    return data;
 }
 
 StringDataPtr StringData::Registry::intern(StringView str)
 {
-    auto it = m_strings.find(str);
-    if (it != m_strings.end())
-        return StringDataPtr{it->value};
-
-    auto data = StringData::create(str);
-    data->refcount |= interned_flag;
-    m_strings.insert({data->strview(), data.get()});
-    return data;
+    return intern(str, hash_value(str));
 }
 
 void StringData::Registry::remove(StringView str)
@@ -43,15 +30,16 @@ void StringData::Registry::remove(StringView str)
 
 void StringData::Registry::debug_stats() const
 {
-    write_to_debug_buffer("Shared Strings stats:");
+    write_to_debug_buffer("Interned Strings stats:");
+    size_t count = m_strings.size();
     size_t total_refcount = 0;
     size_t total_size = 0;
-    size_t count = m_strings.size();
     for (auto& st : m_strings)
     {
-        total_refcount += (st.value->refcount & refcount_mask) - 1;
+        total_refcount += st.value->refcount & refcount_mask;
         total_size += (int)st.value->length;
     }
+    write_to_debug_buffer(format("  count: {}", count));
     write_to_debug_buffer(format("  data size: {}, mean: {}", total_size, (float)total_size/count));
     write_to_debug_buffer(format("  refcounts: {}, mean: {}", total_refcount, (float)total_refcount/count));
 }
