@@ -107,7 +107,7 @@ String replace(StringView str, StringView substr, StringView replacement)
             break;
 
         res += replacement;
-        it = match + (int)substr.length();
+        it = match + substr.length();
     }
     return res;
 }
@@ -215,7 +215,7 @@ InplaceString<23> to_string(float val)
     return to_string_impl<23>(val, std::chars_format::general);
 #else
     InplaceString<23> res;
-    res.m_length = sprintf(res.m_data, "%f", val);
+    res.m_length = snprintf(res.m_data, 23, "%f", val);
     return res;
 #endif
 }
@@ -360,9 +360,15 @@ void format_impl(StringView fmt, ArrayView<const StringView> params, AppendFunc
 
             if (format != closing)
             {
-                for (ColumnCount width = str_to_int({format+1, closing}), len = params[index].column_length();
+                char padding = ' ';
+                if (*(++format) == '0')
+                {
+                    padding = '0';
+                    ++format;
+                }
+                for (ColumnCount width = str_to_int({format, closing}), len = params[index].column_length();
                      width > len; --width)
-                    append(' ');
+                    append(padding);
             }
 
             append(params[index]);
@@ -438,7 +444,7 @@ UnitTest test_string{[]()
     kak_assert(StringView{"youpi"}.ends_with("youpi"));
     kak_assert(not StringView{"youpi"}.ends_with("oup"));
 
-    auto wrapped = "wrap this paragraph\n respecting whitespaces and much_too_long_words" | wrap_at(16) | gather<Vector>();
+    auto wrapped = "wrap this paragraph\n respecting whitespaces and much_too_long_words" | wrap_at(16) | gather<Vector<String>>();
     kak_assert(wrapped.size() == 6);
     kak_assert(wrapped[0] == "wrap this");
     kak_assert(wrapped[1] == "paragraph");
@@ -447,7 +453,7 @@ UnitTest test_string{[]()
     kak_assert(wrapped[4] == "much_too_long_wo");
     kak_assert(wrapped[5] == "rds");
 
-    auto wrapped2 = "error: unknown type" | wrap_at(7) | gather<Vector>();
+    auto wrapped2 = "error: unknown type" | wrap_at(7) | gather<Vector<String>>();
     kak_assert(wrapped2.size() == 3);
     kak_assert(wrapped2[0] == "error:");
     kak_assert(wrapped2[1] == "unknown");
@@ -475,7 +481,7 @@ UnitTest test_string{[]()
     kak_assert(subsequence_match("tchou kanaky", "tchou kanaky"));
     kak_assert(not subsequence_match("tchou kanaky", "tchou  kanaky"));
 
-    kak_assert(format("Youhou {1} {} '{0:4}' \\{}", 10, "hehe", 5) == "Youhou hehe 5 '  10' {}");
+    kak_assert(format("Youhou {1} {} '{0:4}' {2:04} \\{}", 10, "hehe", 5) == "Youhou hehe 5 '  10' 0005 {}");
 
     char buffer[20];
     kak_assert(format_to(buffer, "Hey {}", 15) == "Hey 15");
